import 'CoreLibs/animation.lua'
import 'CoreLibs/animator.lua'
import 'CoreLibs/easing.lua'
import 'CoreLibs/graphics.lua'
import 'CoreLibs/object.lua'
import 'CoreLibs/ui.lua'
import 'utils.lua'

--- @alias Position { x: number, y: number }

--- @alias Base { x: number, y: number }

--- @alias Runner { x: number, y: number, nextBase: Base, prevBase: Base, forcedTo: Base }

--- @alias Fielder { onArrive: fun() | nil, x: number | nil, y: number | nil, target: Position | nil, speed: number }

local gfx = playdate.graphics
playdate.display.setRefreshRate(50)
gfx.setBackgroundColor(gfx.kColorWhite)
playdate.setMenuImage(gfx.image.new("images/game/menu-image.png"))

local grassBackground = gfx.image.new("images/game/grass.png") or {}

local playerImageBlipper = blipper.new(
	100,
    "images/game/player.png",
    "images/game/player-lowhat.png"
)

---@type Position
local backgroundPan = {
    x = 0,
    y = 0,
}

local pitchFlyTimeMs = 2500
local ballStartY, endY = 90, 250
local easingFunction = playdate.easingFunctions.outQuint
local pitchAnimator = gfx.animator.new(pitchFlyTimeMs, ballStartY, endY, easingFunction)

local ballSizeMs = 2000
local ballSizeAnimator = gfx.animator.new(ballSizeMs, 9, 6, playdate.easingFunctions.outBounce)

local screenW, screenH = 400, 240
local centerX, centerY = screenW / 2, screenH / 2
local batBaseX, batBaseY = centerX - 34, 215

local batLength = 45

local tagDistance = 20

local ballY = 999 -- ballStartY
local ballX = 200
local ballSize = 6

local batTipX = 0
local batTipY = 0

local MODES = {
    batting = {},
    running = {}
}
local currentMode = MODES.batting

local hitAnimatorY
local hitAnimatorX

local hitMult = 10

local deltaTime = 0

---@type table<string, Base>
local bases = {
    first = { x = screenW * 0.93, y = screenH * 0.52 },
    second = { x = screenW * 0.47, y = screenH * 0.19 },
    third = { x = screenW * 0.03, y = screenH * 0.52 },
    home = { x = screenW * 0.474, y = screenH * 0.79 }
}

---@type table<Base, Base>
local nextBaseMap = {
    [bases.first] = bases.second,
    [bases.second] = bases.third,
    [bases.third] = bases.home
}

---@type table<string, Fielder>
local fielders = {
    first = {
		speed = 40,
    },
    second = {
		speed = 40,
    },
    shortstop = {
		speed = 40,
    },
    third = {
		speed = 40,
    },
    pitcher = {
		speed = 30,
    },
    catcher = {
		speed = 20,
    },
    left = {
		speed = 40,
    },
    center = {
		speed = 40,
    },
    right = {
		speed = 40,
    }
}

--- Resets the target positions of all fielders to their defaults (at their field positions).
---@param fromOffTheField boolean If provided, also sets all runners' current position to one centralized location.
function resetFielderPositions(fromOffTheField)
	if fromOffTheField then
		for _,fielder in pairs(fielders) do
			fielder.x = centerX
			fielder.y = screenH
		end
	end

    fielders.first.target = { x = screenW - 65, y = screenH * 0.48 }
    fielders.second.target = { x = screenW * 0.70, y = screenH * 0.30 }
    fielders.shortstop.target = { x = screenW * 0.30, y = screenH * 0.30 }
    fielders.third.target = { x = screenW * 0.1, y = screenH * 0.48 }
    fielders.pitcher.target = { x = screenW * 0.48, y = screenH * 0.40 }
    fielders.catcher.target = { x = screenW * 0.475, y = screenH * 0.92 }
    fielders.left.target = { x = screenW * -1, y = screenH * -0.2 }
    fielders.center.target = { x = centerX, y = screenH * -0.4 }
    fielders.right.target = { x = screenW * 2, y = screenH * fielders.left.target.y }
end
resetFielderPositions(true)

local playerStartingX = bases.home.x - 40
local playerStartingY = bases.home.y - 3

--- @type Runner[]
local runners = { }

---@return Runner
function newRunner()
    local new = {
        x = playerStartingX,
        y = playerStartingY,
        nextBase = nil,
        prevBase = nil,
    }
    runners[#runners + 1] = new
    return new
end

---@type Runner | nil
local batter = newRunner()

function throwBall(destX, destY, easingFunc, flyTimeMs)
    if not flyTimeMs then
        flyTimeMs = distanceBetween(ballX, ballY, destX, destY) * 5
    end
    ballSizeAnimator:reset(flyTimeMs)
    hitAnimatorY = gfx.animator.new(flyTimeMs, ballY, destY, easingFunc)
    hitAnimatorX = gfx.animator.new(flyTimeMs, ballX, destX, easingFunc)
end

function pitch()
    pitchAnimator:reset()
    ballY = ballStartY
    ballX = 200
    currentMode = MODES.batting

    backgroundPan.y = 0
    backgroundPan.x = 0
end

function playdate.AButtonDown()
    pitch()
end

function playdate.upButtonDown()
    batBaseY = batBaseY - 1
end

function playdate.downButtonDown()
    batBaseY = batBaseY + 1
end

function playdate.rightButtonDown()
    batBaseX = batBaseX + 1
end

function playdate.leftButtonDown()
    batBaseX = batBaseX - 1
end

local pitchClockSec = 99
local elapsedTime = 0
local crankChange

local baseHitbox = 13
--- Returns the base being touched by the runner at (x,y), or nil, if no base is being touched
function isTouchingBase(x, y)
    for _,base in pairs(bases) do
        if distanceBetween(x, y, base.x, base.y) < baseHitbox then
            return base
        end
    end

    return nil
end

local ballCatchHitbox = 3
function isTouchingBall(x, y)
    local ballDistance = distanceBetween(x, y, ballX, ballY)
    return ballDistance < ballCatchHitbox
end

function updateForcedTos()
end

function outRunner(runnerIndex)
    table.remove(runners, runnerIndex)
    updateForcedTos()
end

function updateFielders()
    local touchingBaseCache = buildCache(
        function(runner)
            return isTouchingBase(runner.x, runner.y)
        end
    )

    for _,fielder in pairs(fielders) do
        -- TODO: Target unforced runners (or their target bases) for tagging
        -- With new Position-based scheme, fielders are now able to set `fielder.target = runner` to track directly
        if fielder.target ~= nil then
            local x, y, distance = normalizeVector(fielder.x, fielder.y, fielder.target.x, fielder.target.y)

            if distance > 1 then
                fielder.x = fielder.x - (x * fielder.speed * deltaTime)
                fielder.y = fielder.y - (y * fielder.speed * deltaTime)
            else
                if fielder.onArrive then
                    fielder.onArrive()
                end
                fielder.target = nil
            end
        end

        if isTouchingBall(fielder.x, fielder.y) then
            local touchedBase = isTouchingBase(fielder.x, fielder.y)
            for i,runner in pairs(runners) do
                local runnerOnBase = touchingBaseCache.get(runner)
                if touchedBase and runner.forcedTo == touchedBase and touchedBase ~= runnerOnBase then
                    outRunner(i)
                elseif not runnerOnBase then
                    local fielderDistance = distanceBetween(runner.x, runner.y, fielder.x, fielder.y)
                    if fielderDistance < tagDistance then
                        outRunner(i)
                    end
                end
            end
        end
    end
end

--- Returns true if at least one runner is still moving
---@return boolean
function updateRunners()
    local autoRunSpeed = 20
    --autoRunSpeed = 140
    local nonPlayerRunners = filter(runners, function (runner)
        return runner ~= batter
    end)

    local runnerMoved = false
    for _,runner in pairs(nonPlayerRunners) do
        local _, nearestBaseDistance = getNearestOf(bases, runner.x, runner.y)
        if runner.nextBase then
            local nb = runner.nextBase
            local x, y, distance = normalizeVector(runner.x, runner.y, nb.x, nb.y)

            if distance > 1 then
                local mult = 1
                if crankChange < 0 then
                    mult = -1
                end
                local prevX, prevY = runner.x, runner.y
                -- TODO: Drift toward nearest base?
                local autoRun = nearestBaseDistance > 5 and mult * autoRunSpeed * deltaTime or 0
                mult = autoRun + (crankChange / 20)
                runner.x = runner.x - (x * mult)
				runner.y = runner.y - (y * mult)
				runnerMoved = runnerMoved or prevX ~= runner.x or prevY ~= runner.y
            else
                runner.nextBase = nextBaseMap[runner.nextBase]
                runner.forcedTo = nil
            end
        end
    end

    return runnerMoved
end

---@return boolean
function ballIsBeingThrown()
    return false
end

---@return boolean
function throwArrivedBeforeRunner()
    return false
end

function getRunnerTargeting(base)
	for _,runner in pairs(runners) do
		if runner.nextBase == base then
			return runner
		end
	end
	return nil
end

---@return Base[]
function getForcedOutTargets()
	local targets = {}
	for _,base in pairs(bases) do
		local runnerTargetingBase = getRunnerTargeting(base)
		if runnerTargetingBase then
		    targets[#targets+1] = base
		else
			return targets
		end
	end
	return targets
    -- return { bases.first }
end

--- Returns the position,distance of the basest closest to the runner furthest from a base
function getBaseOfStrandedRunner()
	local farRunnersBase, farDistance
    for _,runner in pairs(runners) do
    	local base, distance = getNearestOf(bases, runner.x, runner.y, function(base)
			return runner.nextBase == base
    	end)
		if farRunnersBase == nil then
		    farRunnersBase = base
		    farDistance = distance
		elseif farDistance < distance then
			farRunnersBase = base
			farDistance = distance
		end
    end

    return farRunnersBase, farDistance
end

--- Returns x,y of the throw target
---@return number|nil, number|nil
function getNextThrowTarget()
	-- TODO: Handle missed throws, check for fielders at target, etc.
    local targets = getForcedOutTargets()
    print("forcedOutTargets:")
    printTable(targets)
    if #targets ~= 0 then
		return targets[1].x, targets[1].y
	end

	-- local baseCloseToStrandedRunner = getBaseOfStrandedRunner()
	-- return baseCloseToStrandedRunner.x, baseCloseToStrandedRunner.y
end

local resetFieldersAfterSeconds = 5
local secondsSinceLastRunnerMove = 0

function updateGameState()
    deltaTime = playdate.getElapsedTime() or 0
    playdate.resetElapsedTime()
    elapsedTime = elapsedTime + deltaTime

    -- if elapsedTime > pitchClockSec then
    --     elapsedTime = 0
    --     pitch()
    -- end

    if currentMode == MODES.running then
        ballX = hitAnimatorX:currentValue()
        ballY = hitAnimatorY:currentValue()
        ballSize = ballSizeAnimator:currentValue()
    elseif ballY < 999 then
        ballY = pitchAnimator:currentValue()
        ballSize = 6
    end

    local batAngle = math.rad(playdate.getCrankPosition() + 90)
    batTipX = batBaseX + (batLength * math.sin(batAngle))
    batTipY = batBaseY + (batLength * math.cos(batAngle))

    crankChange, acceleratedChange = playdate.getCrankChange()
    if currentMode == MODES.batting and acceleratedChange >= 0 and
      pointDirectlyUnderLine(ballX, ballY, batBaseX, batBaseY, batTipX, batTipY, screenH) then
        currentMode = MODES.running
        ballAngle = batAngle + math.rad(90)

        local mult = math.abs(acceleratedChange / 15)
        local ballVelX = mult * 10 * math.sin(ballAngle)
        local ballVelY = mult * 5 * math.cos(ballAngle)
        if ballVelY > 0 then
            ballVelX = ballVelX * -1
            ballVelY = ballVelY * -1
        end
        ballDestX = ballX + (ballVelX * hitMult)
        ballDestY = ballY + (ballVelY * hitMult)
        throwBall(ballDestX, ballDestY, playdate.easingFunctions.outQuint, 2000)

        fielders.first.target = bases.first
        batter.nextBase = bases.first
        batter.forcedTo = bases.first
        batter = nil -- Demote batter to a mere runner

        local chasingFielder = getNearestOf(fielders, ballDestX, ballDestY)
        chasingFielder.target = { x = ballDestX, y = ballDestY }
        chasingFielder.onArrive = function()
            local targetX, targetY = getNextThrowTarget()
            if targetX ~= nil then
				throwBall(targetX, targetY, playdate.easingFunctions.linear)
				chasingFielder.onArrive = nil
            end
        end
    end

    if currentMode == MODES.running then
        if updateRunners() then
            secondsSinceLastRunnerMove = 0
	    else
            secondsSinceLastRunnerMove = secondsSinceLastRunnerMove + deltaTime
            if secondsSinceLastRunnerMove > resetFieldersAfterSeconds then
                resetFielderPositions(false)
                currentMode = MODES.batting
                batter = newRunner()
            end
        end
    end
	updateFielders()
end

function playdate.update()
    updateGameState()
    playdate.graphics.animation.blinker.updateAll()

    -- TODO: Show baserunning minimap when panning?
    local ballBuffer = 5
    if ballY < ballBuffer then
        backgroundPan.y = math.max(ballBuffer, -1 * (ballY - ballBuffer))
    else
        backgroundPan.y = 0
    end
    if ballX < ballBuffer then
        backgroundPan.x = math.max(-400, -1 * (ballX - ballBuffer))
    elseif ballX > (screenW - ballBuffer) then
        backgroundPan.x = math.min(800, -1 * (ballX - ballBuffer))
    end

    if ballX > 0 and ballX < (screenW - ballBuffer) then
        backgroundPan.x = 0
    end

    gfx.clear()
    grassBackground:draw(backgroundPan.x - 400, backgroundPan.y - 240)

    gfx.setColor(gfx.kColorBlack)
    gfx.setLineWidth(2)
    gfx.drawCircleAtPoint(ballX + backgroundPan.x, ballY + backgroundPan.y, ballSize)

    for _,fielder in pairs(fielders) do
        gfx.fillRect(fielder.x + backgroundPan.x, fielder.y + backgroundPan.y, 14, 25)
    end

    gfx.setLineWidth(5)

    if currentMode == MODES.batting then
        gfx.drawLine(
            batBaseX + backgroundPan.x, batBaseY + backgroundPan.y,
            batTipX + backgroundPan.x, batTipY + backgroundPan.y
        )
    end

    if playdate.isCrankDocked() or (crankChange < 2 and currentMode == MODES.running) then
        playdate.ui.crankIndicator:draw()
    end

    -- TODO? Change blip speed depending on runner speed?
    for _,runner in pairs(runners) do
		-- TODO? Scale sprites down as y increases
        playerImageBlipper:draw(
                false,
                runner.x + backgroundPan.x,
                runner.y + backgroundPan.y
        )
    end

end
